import numpy as np
import heapq
import scipy.signal as signal
from scipy import ndimage

#Based on https://github.com/rrlyman/phase-reconstruction

def dxdw(x):
    ''' return the derivative of x with respect to frequency'''
    xp = np.pad(x,1,mode='edge')
    dw = (xp[1:-1,2:]-xp[1:-1,:-2])/2

    return dw
    
def dxdt(x):
    ''' return the derivative of x with respect to time'''   
    xp = np.pad(x,1,mode='edge')      
    dt = (xp[2:,1:-1]-xp[:-2,1:-1])/(2)

    return dt

def calculate_synthesis_window(win_length=2048, hop_length=512, n_fft=2048,window=None):
    gsynth = np.zeros_like(window)
    for l in range(int(n_fft)):
        denom = 0                
        for n in range(-win_length//hop_length, win_length//hop_length+1):
            dl = l-n*hop_length
            if dl >=0 and dl < win_length:
                denom += window[dl]**2
            gsynth[l] = window[l]/denom

    return gsynth

def get_default_window(n_fft):
    lambda_ = (-n_fft**2/(8*np.log(0.01)))**.5
    lambdasqr = lambda_**2
    gamma = 2*np.pi*lambdasqr
    window=np.array(signal.windows.gaussian(2*n_fft+1, lambda_*2, sym=False))[1:2*n_fft+1:2]

    return window, gamma, lambdasqr

def pghi(x, win_length=2048, hop_length=512, n_fft=None, window=None, synthesis_window=None,gamma=None,lambdasqr=None, tol=1e-6):
    """
    Performs phase gradient heap integration to estimate the phase of a spectrogram given the magnitude
    x:           magnitude spectrogram
    win_length:  size (in samples) of the analysis window
    hop_length:  number of samples between two consecutive windows
    n_fft:       (default = win_length), number of samples to calculate the FFT. If it is larger than win_length, window will get padded
    window:      (default is a gaussian window, it is recommended to use it). Array of size win_length.
    synthesis_window: window to use for the synthesis step. By default it will be calculated from window
    gamma:       constant related to the window (if not given, is calculated from lambdasqr)
    lambdasqr:   constant related to the window (if not given, the default value corresponds to the gaussian window)
    tol:         small signal relative magnitude filtering size
    """
    if n_fft is None: n_fft = win_length
    if gamma is not None: lambdasqr = gamma/(2*np.pi)
    if window is None:
        window, gamma, lambdasqr = get_default_window(n_fft)
    if synthesis_window is None:
        gsynth = calculate_synthesis_window(win_length, hop_length, n_fft, window)
    
    m2 = int(win_length/2) + 1
    n_frames = x.shape[0]
    wbin = 2*np.pi/win_length
    logs = np.log(x+1e-50)
    fmul = gamma/(hop_length * win_length)

    tgrad = dxdw(logs)/fmul + (2*np.pi*hop_length/win_length)*np.arange(m2)
    fgrad = -fmul*dxdt(logs) + np.pi

    dphaseNE =  tgrad  + fgrad
    dphaseSE =  tgrad - fgrad
    phase = np.random.random_sample(x.shape)*2*np.pi

    mask = x > (tol*np.max(x))
    active_padded = np.pad(mask,1,mode='constant',constant_values=False)

    while np.any(active_padded):
        labeled, nr_objects = ndimage.label(active_padded[1:n_frames+1,1:m2+1])
        startpoints = ndimage.maximum_position(x*active_padded[1:n_frames+1,1:m2+1], labeled, range(nr_objects+1))[1:]
        for startpoint in startpoints:
            h=[]
            n0,m0 = startpoint
            phase[n0,m0] = 0
            heapq.heappush(h, (-x[n0,m0],n0,m0)) 
            active_padded[n0+1,m0+1]=False
            while len(h) > 0:
                s=heapq.heappop(h)            
                n,m = s[1],s[2]
                if active_padded[(n+1), m+2]: # North             
                    active_padded[(n+1), m+2]=False # padded is 1 indexed                               
                    phase[n, m+1]=  phase[n,  m] +(fgrad[n,  m] + fgrad[n, m+1])/2 
                    heapq.heappush(h, (-x[n, m+1],n,m+1))                                       
                if active_padded[(n+1), m]: # South 
                    active_padded[(n+1), m]=False # padded is 1 indexed                             
                    phase[n, m-1]=  phase[n,  m] - (fgrad[n,  m] + fgrad[n, m-1])/2 
                    heapq.heappush(h, (-x[n, m-1],n,m-1))    
                                                          
                if active_padded[(n+2), m+1]:  # East 
                    active_padded[(n+2), m+1]=False # padded is 1 indexed                          
                    phase[(n+1), m]=  phase[n,  m] + (tgrad[n,  m] + tgrad[(n+1), m])/2 
                    heapq.heappush(h, (-x[(n+1), m], n+1,m))   
                                                                      
                if active_padded[n, m+1]: # West            
                    active_padded[n, m+1]=False # padded is 1 indexed                             
                    phase[(n-1), m]=  phase[n,  m] - (tgrad[n,  m] + tgrad[(n-1), m])/2 
                    heapq.heappush(h, (-x[(n-1), m],n-1,m))
                                                                 
    return phase

def stft(x,win_length=2048,hop_length=512,window=None):
    """
    Performs the short fourier transform of x in a way that is compatible with PGHI
    x:           time domain signal
    win_length:  size (in samples) of the analysis window
    hop_length:  number of samples between two consecutive windows
    window:      array of size win_length with the window to apply
    """
    L = x.shape[0] - win_length
    if window is None:
        window, _, _ = get_default_window(win_length)
    return np.stack([np.fft.rfft(window*x[ix:ix + win_length]) for ix in range(0, L, hop_length)])

def istft(X, win_length=2048,hop_length=512,window=None,synthesis_window=None):
    """
    Returns a spectrogram to the time domain by using the Overlap and Add method
    X:                spectrogram
    win_length:       size (in samples) of the analysis window
    hop_length:       number of samples between two consecutive windows
    window:           array of size win_length with the window to apply
    synthesis_window: window to use for the synthesis step. By default it will be calculated from window
    """
    N = X.shape[0]
    vr=np.fft.irfft(X)
    sig = np.zeros((N*hop_length+win_length))

    if window is None:
        window, _, _ = get_default_window(win_length)
    if synthesis_window is None:
        synthesis_window = calculate_synthesis_window(win_length, hop_length, win_length, window)

    for n in range(N):
        vs = vr[n]*synthesis_window
        sig[n*hop_length: n*hop_length+win_length] += vs
    
    return sig

def griffin_lim(X, win_length=2048,hop_length=512,window=None,synthesis_window=None, n_iters=100):
    """
    An implementation of the Griffin Lim algorithm
    X:                spectrogram (can be complex and have an initial phase)
    win_length:       size (in samples) of the analysis window
    hop_length:       number of samples between two consecutive windows
    window:           array of size win_length with the window to apply
    synthesis_window: window to use for the synthesis step. By default it will be calculated from window
    n_iters:          number of iterations to perform
    """
    if window is None:
        window, _, _ = get_default_window(win_length)
    if synthesis_window is None:
        synthesis_window = calculate_synthesis_window(win_length, hop_length, win_length, window)
    
    mag_X = np.abs(X)
    for i in range(n_iters):
        x = istft(X, win_length, hop_length, window, synthesis_window)
        X = stft(x,win_length,hop_length,window)
        X = mag_X*np.exp(1.0j*np.angle(X))
    return x


from pathlib import Path
import requests
import pafy
import subprocess
import paramiko
import time
import boto3
import copy

def download_url(url,download_path):
    """
    Download an url
    url:           url to download
    download_path: path where to save the downloaded url
    """
    r = requests.get(url, stream = True)
    if not Path(download_path).exists():
        print("Downloading file {}".format(url)) 
        with open(download_path,"wb") as dest_file: 
            for chunk in r.iter_content(chunk_size=1024): 
                if chunk: 
                    dest_file.write(chunk)
        return 1
    else:
        return 0

def download_audio_from_yt(url_id,start=None,end=None,download_path=None):
    """
    Download the audio from a youtube video
    url_id:         url of the video to download
    start:          start (in seconds) of the video fragment to download
    end:            end (in seconds) of the video fragment to download
    download_path:  path where to save the downloaded audio
    """
    video_page_url='https://www.youtube.com/watch?v={}'.format(url_id)
    #Obtengo la URL del archivo de video con mejor audio:
    video = pafy.new(video_page_url)
    video_duration = video.length
    best_audio = video.getbestaudio().url
    #Descargo la parte deseada usando ffmpeg y la guardo en un mkv sin reencodear
    cmd = ['ffmpeg','-i',best_audio,'-vn','-ss','{}'.format(int(start)),'-to','{}'.format(int(end)),'-acodec','copy','temp_out.mkv']
    subprocess.call(cmd,timeout=15)
    if Path('temp_out.mkv').exists():
        return 'temp_out.mkv'
    else:
        return None

def run_ssh_commands(username,hostname,pem_file,command):
    ssh_client = paramiko.SSHClient()
    ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    privkey = paramiko.RSAKey.from_private_key_file(pem_file)
    if Path(command).exists():
        with open(command,'r') as f:
            command = f.read()

    failed = True
    while failed:   
        try:
            ssh_client.connect(hostname=hostname,username=username,pkey=privkey)
            failed = False
        except:
            failed = True
            time.sleep(1)
            print('Retrying connection')

    stdin, stdout, stderr = ssh_client.exec_command(command)
    ssh_client.close()
    return stdin, stdout, stderr

def run_ssh_commands_2(username,hostname,command,pem_file=None,ec2_instance_connect_data=None):
    if ec2_instance_connect_data:
        ec2_instance_connect_data = copy.deepcopy(ec2_instance_connect_data)
        client = boto3.client('ec2-instance-connect',ec2_instance_connect_data['region'])
        ec2_instance_connect_data.pop('region')
        client.send_ssh_public_key(**ec2_instance_connect_data)

    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    privkey = paramiko.RSAKey.from_private_key_file(pem_file)
    if Path(command).exists():
        with open(command,'r') as f:
            command = f.read()

    failed = True
    while failed:
        try:
            ssh.connect(hostname, username=username, pkey=privkey)
            sleeptime = 0.1
            outdata, errdata = '', ''
            ssh_transp = ssh.get_transport()
            chan = ssh_transp.open_session()
            chan.setblocking(0)
            chan.exec_command(command)
            while True:  # monitoring process
                # Reading from output streams
                while chan.recv_ready():
                    outdata += chan.recv(1000).decode('utf-8')
                while chan.recv_stderr_ready():
                    errdata += chan.recv_stderr(1000).decode('utf-8')
                if chan.exit_status_ready():  # If completed
                    break
                time.sleep(sleeptime)
            retcode = chan.recv_exit_status()
            ssh_transp.close()
            failed = False
        except:
            failed = True
            time.sleep(1)
            print('Retrying connection')
    
    return outdata, errdata


import boto3
import tqdm
from pathlib import Path

def get_instances_info(region='us-east-1',input_credentials=False):
    """
    Returns a list of dictionaries, with each dictionary having information about the instances
    running in a region. The dictionary keys are the id, name, PublicDnsName and PrivateIpAddress 
    of the instance.

    region:            region in which to look for instances
    input_credentials: if True, the user will be asked for the access_key_id and secret_access_key
    """
    if input_credentials:       
        key_id = input('aws_access_key_id')
        secret_key = input('aws_secret_access_key')
        ec2 = boto3.client('ec2',region,aws_access_key_id=key_id,aws_secret_access_key=secret_key)
    else:
        ec2 = boto3.client('ec2',region)

    response = ec2.describe_instances()
    info_instances = []
    for group in response['Reservations']: 
        for instance in group['Instances']: 
            instance_i = { 
                'id': instance['InstanceId'], 
                'name': [tag['Value'] for tag in instance['Tags'] if tag['Key']=='Name'][0], 
                'PublicDnsName': instance['PublicDnsName']
                } 
            if 'PrivateIpAddress' in instance:
                instance_i['PrivateIpAddress'] = instance['PrivateIpAddress']
            info_instances.append(instance_i)

    return info_instances

def get_all_s3_objects(s3, **base_kwargs):
    """
    Based on:
    https://stackoverflow.com/questions/54314563/how-to-get-more-than-1000-objects-from-s3-by-using-list-objects-v2/54314628
    """
    continuation_token = None
    while True:
        list_kwargs = dict(MaxKeys=1000, **base_kwargs)
        if continuation_token:
            list_kwargs['ContinuationToken'] = continuation_token
        response = s3.list_objects_v2(**list_kwargs)
        yield from response.get('Contents', [])
        if not response.get('IsTruncated'):  # At the end of the list?
            break
        continuation_token = response.get('NextContinuationToken')

def download_s3(bucket_name, bucket_path, download_path, exclude=None, if_exists='abort',input_credentials=False):
    """
    Downloads S3 objects

    bucket_name:       name of the S3 bucket
    bucket_path:       path of the S3 object
    download_path:     destination path of the downloaded objects
    exclude:           list of objects to exclude from download
    if_exists:         behaviour when a file to download already exists in the download_path
                       can be 'abort' and the file won't be downloaded
                       can be 'replace' and the file will be downloaded anyway and replace the existing file
    input_credentials: if True, the user will be asked for the access_key_id and secret_access_key
                 
    """
    if input_credentials:
        key_id = input('aws_access_key_id')
        secret_key = input('aws_secret_access_key')
        s3_client = boto3.client('s3',aws_access_key_id=key_id,aws_secret_access_key=secret_key)
    else:
        s3_client = boto3.client('s3')

    if exclude is None:
        exclude = []

    #all_keys = s3_client.list_objects_v2(Bucket=bucket_name,Prefix=bucket_path)
    all_keys = [f for f in get_all_s3_objects(s3_client, Bucket=bucket_name, Prefix=bucket_path)]
    keys = [k['Key'] for k in all_keys if k['Key'] not in exclude and k['Key'][-1] != '/']

    download_path = Path(download_path)

    if not download_path.parent.exists():
        download_path.parent.mkdir(parents=True)

    for key in tqdm.tqdm(keys):
        relative_key = str(Path(key).relative_to(bucket_path))
        destination_path = Path(download_path,relative_key)
        download = False
        if destination_path.exists():
            if if_exists == 'abort':
                print('Skipping {} as it already exists'.format(key))
            elif if_exists == 'replace':
                download = True
            else:
                raise Exception('Unknown if_exists value {}'.format(if_exists))
        else:
            download = True
            if not destination_path.parent.exists():
                destination_path.parent.mkdir(parents=True)

        if download:
            print('Downloading {}'.format(key))
            s3_client.download_file(Bucket=bucket_name,Key=key,Filename=str(destination_path))

class S3File:
    """
    Class to manage S3 objects with an interface similar to pathlib Path
    """
    def __init__(self,*args):
        """
        Can receive a list of strings, S3Files or Paths, indicating path parts
        """
        args = list(args)
        for i, arg in enumerate(args):
            if isinstance(arg,S3File):
                arg = str(arg)

            if isinstance(arg,str):
                if i == 0 and arg.startswith('s3:'):
                    if arg.startswith('s3://'):
                        args[i] = arg
                    elif arg.startswith('s3:/'):
                        args[i] = 's3://' + '/'.join(arg.split('s3:/')[1:])
                    elif arg.startswith('s3:'):
                        args[i] = 's3://' + '/'.join(arg.split('s3:')[1:])
                else:
                    args[i] = arg
            elif isinstance(arg,Path):
                arg_parts = list(arg.parts)
                if arg.parts[0].startswith('s3:'):
                    arg_parts[0] = 's3:/'
                    args[i] = '/'.join(arg_parts)
                else:
                    args[i] = 's3://' + str(arg)

        self.path = '/'.join(args)
        if not self.path.startswith('s3://'):
            self.path = 's3://' + self.path
        self.s3_client = boto3.client('s3')

    def __repr__(self):
        return self.path

    def download(self,download_path):
        """
        Downloads the current S3 object into download_path
        """
        if isinstance(download_path,str):
            download_path = Path(download_path)
        download_path = download_path.expanduser()
        if not download_path.parent.exists():
            download_path.parent.mkdir(parents=True)
        self.s3_client.download_file(Bucket=self.get_bucket_name(),Key=self.get_key(),Filename=str(download_path))
    
    def upload(self,source_path):
        """
        Uploads the file in source_path to the S3 path
        """
        self.s3_client.upload_file(str(source_path),self.get_bucket_name(),self.get_key())

    def exists(self):
        """
        Checks if the S3 object exists
        """
        bucket_name = self.get_bucket_name()
        key = self.path.split('s3://{}/'.format(bucket_name))[-1]

        results = self.s3_client.list_objects_v2(Bucket=bucket_name,Prefix=key)
        
        #all_keys = ['s3://{}/{}'.format(bucket_name,k['Key']) for k in get_all_s3_objects(self.s3_client, Bucket=bucket_name)]

        #return self.path in all_keys
        return results['KeyCount']>0

    def get_bucket_name(self):
        """
        Gets the name of the bucket the S3 path refers to.
        """
        return self.path.split('s3://')[-1].split('/')[0]

    def get_key(self):
        """
        Gets the key of the S3 object.
        """
        return '/'.join(self.path.split('s3://')[-1].split('/')[1:])

    def glob(self,pattern):
        """
        Find all objects matching a glob pattern in the S3 bucket
        """
        from pathlib import PurePath
        bucket_name = self.get_bucket_name()

        all_keys = ['s3://{}/{}'.format(bucket_name,k['Key']) for k in get_all_s3_objects(self.s3_client, Bucket=bucket_name)]
        match_keys = [k for k in all_keys if PurePath(k).match(pattern)]

        return match_keys

    def delete(self):
        """
        Deletes the object from S3
        """
        self.s3_client.delete_object(Bucket=self.get_bucket_name(), Key=self.get_key())

    def __getstate__(self):
        d = self.__dict__
        if 's3_client' in d:
            d.pop('s3_client')
        return d

    def __setstate__(self,d):
        d['s3_client'] = boto3.client('s3')
        self.__dict__ = d
        




import subprocess
import shlex

def run_command(cmd,silent=False,timeout=15):
    """
    Runs a command in a terminal
    cmd:     command to run
    silent:  if True, logs are not shown
    timeout: integer telling the number of seconds to wait before aborting the command
    """
    if isinstance(cmd,str):
        cmd = shlex.split(cmd)

    if silent:
        subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, timeout=timeout)
    else:
        subprocess.check_call(cmd,timeout=timeout)



from pathlib import Path
import tqdm

class CompressedFile:
    """
    Class to wrap compressed files
    """
    def __init__(self,path):
        """
        path: path to the compressed file
        """
        self.file = Path(path).expanduser()
        self.extension = self.file.suffix

    def add_file(self,filename,filename_inside):
        """
        add file to the container (supports .zip)
        filename:        path to the file to include in the container
        filename_inside: name of the file when inside the container
        """
        if self.extension == '.zip':
            zip_file = zipfile.ZipFile(self.file(),'w')
            zip_file.write(filename,arcname=filename_inside)
            zip_file.close()

    def extract(self,destination_path):
        """
        extract files in container (supports .zip and .gz)
        destination_file: path where the files will be extracted
        """
        destination_path = Path(destination_path).expanduser()
        if not destination_path.exists():
            destination_path.mkdir(parents=True)
        if self.extension == '.gz':
            import tarfile
            tar = tarfile.open(self.file)
            tar_members = tar.getnames()
            for member in tqdm.tqdm(tar_members):
                if not Path(destination_path,member).exists():
                    tar.extract(member,destination_path)
            tar.close()
        elif self.extension == '.zip':
            import zipfile
            zip_file = zipfile.ZipFile(self.file)
            zip_members = zip_file.namelist()
            for member in tqdm.tqdm(zip_members):
                if not Path(destination_path,member).exists():
                    zip_file.extract(member,destination_path)
            zip_file.close()


from pyknife.console import run_command

def ffmpeg_to_wav(in_file,out_file,sr=None):
    """
    Uses ffmpeg to turn any audio file into a .wav file
    in_file:  path to the audio to convert
    out_file: destination path of the converted audio
    sr:       desired sample rate of the converted audio
    """
    if sr is None:
        cmd = "ffmpeg -y -i {} {}".format(in_file,sr,out_file)
    else:
        cmd = "ffmpeg -y -i {} -ar {} {}".format(in_file,sr,out_file)
    run_command(cmd,silent=True)